@@ -12,20 +12,24 @@
 # individuals. For the exact contribution history, see the revision
 # history and logs, available at http://trac.edgewall.org/log/.
 
+from __future__ import with_statement
+
 from ConfigParser import ConfigParser
 from copy import deepcopy
 import os.path
 
+from genshi.builder import tag
 from trac.admin import AdminCommandError, IAdminCommandProvider
 from trac.core import *
 from trac.util import AtomicFile, as_bool
-from trac.util.compat import any
+from trac.util.compat import cleandoc, wait_for_file_mtime_change
 from trac.util.text import printout, to_unicode, CRLF
-from trac.util.translation import _, N_
+from trac.util.translation import _, N_, tag_
 
-__all__ = ['Configuration', 'Option', 'BoolOption', 'IntOption', 'FloatOption',
-           'ListOption', 'ChoiceOption', 'PathOption', 'ExtensionOption',
-           'OrderedExtensionsOption', 'ConfigurationError']
+__all__ = ['Configuration', 'ConfigSection', 'Option', 'BoolOption',
+           'IntOption', 'FloatOption', 'ListOption', 'ChoiceOption',
+           'PathOption', 'ExtensionOption', 'OrderedExtensionsOption',
+           'ConfigurationError']
 
 # Retained for backward-compatibility, use as_bool() instead
 _TRUE_VALUES = ('yes', 'true', 'enabled', 'on', 'aye', '1', 1, True)
@@ -40,6 +44,12 @@ class ConfigurationError(TracError):
     """Exception raised when a value in the configuration file is not valid."""
     title = N_('Configuration Error')
 
+    def __init__(self, message=None, title=None, show_traceback=False):
+        if message is None:
+            message = _("Look in the Trac log for more information.")
+        super(ConfigurationError, self).__init__(message, title,
+                                                 show_traceback)
+
 
 class Configuration(object):
     """Thin layer over `ConfigParser` from the Python standard library.
@@ -48,7 +58,7 @@ class Configuration(object):
     the last modification time of the configuration file, and reparses it
     when the file has changed.
     """
-    def __init__(self, filename):
+    def __init__(self, filename, params={}):
         self.filename = filename
         self.parser = ConfigParser()
         self._old_sections = {}
@@ -74,43 +84,43 @@ class Configuration(object):
 
     def get(self, section, key, default=''):
         """Return the value of the specified option.
-        
+
         Valid default input is a string. Returns a string.
         """
         return self[section].get(key, default)
 
     def getbool(self, section, key, default=''):
         """Return the specified option as boolean value.
-        
+
         If the value of the option is one of "yes", "true", "enabled", "on",
         or "1", this method wll return `True`, otherwise `False`.
-        
+
         Valid default input is a string or a bool. Returns a bool.
-        
+
         (since Trac 0.9.3, "enabled" added in 0.11)
         """
         return self[section].getbool(key, default)
 
     def getint(self, section, key, default=''):
         """Return the value of the specified option as integer.
-        
+
         If the specified option can not be converted to an integer, a
         `ConfigurationError` exception is raised.
-        
+
         Valid default input is a string or an int. Returns an int.
-        
+
         (since Trac 0.10)
         """
         return self[section].getint(key, default)
 
     def getfloat(self, section, key, default=''):
         """Return the value of the specified option as float.
-        
+
         If the specified option can not be converted to a float, a
         `ConfigurationError` exception is raised.
-        
+
         Valid default input is a string, float or int. Returns a float.
-        
+
         (since Trac 0.12)
         """
         return self[section].getfloat(key, default)
@@ -118,23 +128,23 @@ class Configuration(object):
     def getlist(self, section, key, default='', sep=',', keep_empty=False):
         """Return a list of values that have been specified as a single
         comma-separated option.
-        
+
         A different separator can be specified using the `sep` parameter. If
         the `keep_empty` parameter is set to `True`, empty elements are
         included in the list.
-        
+
         Valid default input is a string or a list. Returns a string.
-        
+
         (since Trac 0.10)
         """
         return self[section].getlist(key, default, sep, keep_empty)
 
     def getpath(self, section, key, default=''):
         """Return a configuration value as an absolute path.
-        
+
         Relative paths are resolved relative to the location of this
         configuration file.
-        
+
         Valid default input is a string. Returns a normalized path.
 
         (enabled since Trac 0.11.5)
@@ -143,7 +153,7 @@ class Configuration(object):
 
     def set(self, section, key, value):
         """Change a configuration value.
-        
+
         These changes are not persistent unless saved with `save()`.
         """
         self[section].set(key, value)
@@ -151,7 +161,7 @@ class Configuration(object):
     def defaults(self, compmgr=None):
         """Returns a dictionary of the default configuration values
         (''since 0.10'').
-        
+
         If `compmgr` is specified, return only options declared in components
         that are enabled in the given `ComponentManager`.
         """
@@ -163,7 +173,7 @@ class Configuration(object):
     def options(self, section, compmgr=None):
         """Return a list of `(name, value)` tuples for every option in the
         specified section.
-        
+
         This includes options that have default values that haven't been
         overridden. If `compmgr` is specified, only return default option
         values for components that are enabled in the given `ComponentManager`.
@@ -176,7 +186,7 @@ class Configuration(object):
 
     def sections(self, compmgr=None, defaults=True):
         """Return a list of section names.
-        
+
         If `compmgr` is specified, only the section names corresponding to
         options declared in components that are enabled in the given
         `ComponentManager` are returned.
@@ -192,7 +202,7 @@ class Configuration(object):
         """Returns True if option exists in section in either the project
         trac.ini or one of the parents, or is available through the Option
         registry.
-        
+
         (since Trac 0.11)
         """
         section_str = _to_utf8(section)
@@ -231,11 +241,12 @@ class Configuration(object):
 
         # At this point, all the strings in `sections` are UTF-8 encoded `str`
         try:
-            fileobj = AtomicFile(self.filename, 'w')
-            try:
+            wait_for_file_mtime_change(self.filename)
+            with AtomicFile(self.filename, 'w') as fileobj:
                 fileobj.write('# -*- coding: utf-8 -*-\n\n')
-                for section, options in sections:
-                    fileobj.write('[%s]\n' % section)
+                for section_str, options in sections:
+                    fileobj.write('[%s]\n' % section_str)
+                    section = to_unicode(section_str)
                     for key_str, val_str in options:
                         if to_unicode(key_str) in self[section].overridden:
                             fileobj.write('# %s = <inherited>\n' % key_str)
@@ -244,8 +255,6 @@ class Configuration(object):
                                              .replace('\n', '\n ')
                             fileobj.write('%s = %s\n' % (key_str, val_str))
                     fileobj.write('\n')
-            finally:
-                fileobj.close()
             self._old_sections = deepcopy(self.parser._sections)
         except Exception:
             # Revert all changes to avoid inconsistencies
@@ -259,7 +268,6 @@ class Configuration(object):
         changed = False
         modtime = os.path.getmtime(self.filename)
         if force or modtime > self._lastmtime:
-            self._sections = {}
             self.parser._sections = {}
             if not self.parser.read(self.filename):
                 raise TracError(_("Error reading '%(file)s', make sure it is "
@@ -267,7 +275,7 @@ class Configuration(object):
             self._lastmtime = modtime
             self._old_sections = deepcopy(self.parser._sections)
             changed = True
-        
+
         if changed:
             self.parents = []
             if self.parser.has_option('inherit', 'file'):
@@ -280,35 +288,37 @@ class Configuration(object):
         else:
             for parent in self.parents:
                 changed |= parent.parse_if_needed(force=force)
-        
+
         if changed:
-            self._cache = {}
+            self._sections = {}
         return changed
 
     def touch(self):
         if self.filename and os.path.isfile(self.filename) \
-           and os.access(self.filename, os.W_OK):
+                and os.access(self.filename, os.W_OK):
+            wait_for_file_mtime_change(self.filename)
             os.utime(self.filename, None)
 
     def set_defaults(self, compmgr=None):
         """Retrieve all default values and store them explicitly in the
         configuration, so that they can be saved to file.
-        
+
         Values already set in the configuration are not overridden.
         """
-        for section, default_options in self.defaults(compmgr).items():
-            for name, value in default_options.items():
-                if not self.parser.has_option(_to_utf8(section),
-                                              _to_utf8(name)):
-                    if any(parent[section].contains(name, defaults=False)
-                           for parent in self.parents):
-                        value = None
-                    self.set(section, name, value)
+        for (section, name), option in Option.get_registry(compmgr).items():
+            if not self.parser.has_option(_to_utf8(section), _to_utf8(name)):
+                value = option.default
+                if any(parent[section].contains(name, defaults=False)
+                       for parent in self.parents):
+                    value = None
+                if value is not None:
+                    value = option.dumps(value)
+                self.set(section, name, value)
 
 
 class Section(object):
     """Proxy for a specific configuration section.
-    
+
     Objects of this class should not be instantiated directly.
     """
     __slots__ = ['config', 'name', 'overridden', '_cache']
@@ -325,13 +335,13 @@ class Section(object):
         for parent in self.config.parents:
             if parent[self.name].contains(key, defaults=False):
                 return True
-        return defaults and Option.registry.has_key((self.name, key))
-    
+        return defaults and (self.name, key) in Option.registry
+
     __contains__ = contains
 
     def iterate(self, compmgr=None, defaults=True):
         """Iterate over the options in this section.
-        
+
         If `compmgr` is specified, only return default option values for
         components that are enabled in the given `ComponentManager`.
         """
@@ -354,13 +364,13 @@ class Section(object):
                     yield option
 
     __iter__ = iterate
-    
+
     def __repr__(self):
-        return '<Section [%s]>' % (self.name)
+        return '<%s [%s]>' % (self.__class__.__name__, self.name)
 
     def get(self, key, default=''):
         """Return the value of the specified option.
-        
+
         Valid default input is a string. Returns a string.
         """
         cached = self._cache.get(key, _use_default)
@@ -378,7 +388,7 @@ class Section(object):
             else:
                 if default is not _use_default:
                     option = Option.registry.get((self.name, key))
-                    value = option and option.default or _use_default
+                    value = option.default if option else _use_default
                 else:
                     value = _use_default
         if value is _use_default:
@@ -392,7 +402,7 @@ class Section(object):
 
     def getbool(self, key, default=''):
         """Return the value of the specified option as boolean.
-        
+
         This method returns `True` if the option value is one of "yes", "true",
         "enabled", "on", or non-zero numbers, ignoring case. Otherwise `False`
         is returned.
@@ -403,10 +413,10 @@ class Section(object):
 
     def getint(self, key, default=''):
         """Return the value of the specified option as integer.
-        
+
         If the specified option can not be converted to an integer, a
         `ConfigurationError` exception is raised.
-        
+
         Valid default input is a string or an int. Returns an int.
         """
         value = self.get(key, default)
@@ -421,10 +431,10 @@ class Section(object):
 
     def getfloat(self, key, default=''):
         """Return the value of the specified option as float.
-        
+
         If the specified option can not be converted to a float, a
         `ConfigurationError` exception is raised.
-        
+
         Valid default input is a string, float or int. Returns a float.
         """
         value = self.get(key, default)
@@ -440,11 +450,11 @@ class Section(object):
     def getlist(self, key, default='', sep=',', keep_empty=True):
         """Return a list of values that have been specified as a single
         comma-separated option.
-        
+
         A different separator can be specified using the `sep` parameter. If
         the `keep_empty` parameter is set to `False`, empty elements are omitted
         from the list.
-        
+
         Valid default input is a string or a list. Returns a list.
         """
         value = self.get(key, default)
@@ -455,7 +465,7 @@ class Section(object):
         else:
             items = list(value)
         if not keep_empty:
-            items = filter(None, items)
+            items = [item for item in items if item not in (None, '')]
         return items
 
     def getpath(self, key, default=''):
@@ -473,7 +483,7 @@ class Section(object):
 
     def options(self, compmgr=None):
         """Return `(key, value)` tuples for every option in the section.
-        
+
         This includes options that have default values that haven't been
         overridden. If `compmgr` is specified, only return default option
         values for components that are enabled in the given `ComponentManager`.
@@ -483,7 +493,7 @@ class Section(object):
 
     def set(self, key, value):
         """Change a configuration value.
-        
+
         These changes are not persistent unless saved with `save()`.
         """
         self._cache.pop(key, None)
@@ -509,37 +519,82 @@ class Section(object):
             self.config.parser.remove_option(_to_utf8(self.name), _to_utf8(key))
 
 
-class Option(object):
-    """Descriptor for configuration options on `Configurable` subclasses."""
+def _get_registry(cls, compmgr=None):
+    """Return the descriptor registry.
+
+    If `compmgr` is specified, only return descriptors for components that
+    are enabled in the given `ComponentManager`.
+    """
+    if compmgr is None:
+        return cls.registry
+
+    from trac.core import ComponentMeta
+    components = {}
+    for comp in ComponentMeta._components:
+        for attr in comp.__dict__.itervalues():
+            if isinstance(attr, cls):
+                components[attr] = comp
+
+    return dict(each for each in cls.registry.iteritems()
+                if each[1] not in components
+                   or compmgr.is_enabled(components[each[1]]))
+
+
+class ConfigSection(object):
+    """Descriptor for configuration sections."""
 
     registry = {}
-    accessor = Section.get
+
+    @staticmethod
+    def get_registry(compmgr=None):
+        """Return the section registry, as a `dict` mapping section names to
+        `ConfigSection` objects.
+
+        If `compmgr` is specified, only return sections for components that are
+        enabled in the given `ComponentManager`.
+        """
+        return _get_registry(ConfigSection, compmgr)
+
+    def __init__(self, name, doc, doc_domain='tracini'):
+        """Create the configuration section."""
+        self.name = name
+        self.registry[self.name] = self
+        self.__doc__ = cleandoc(doc)
+        self.doc_domain = doc_domain
+
+    def __get__(self, instance, owner):
+        if instance is None:
+            return self
+        config = getattr(instance, 'config', None)
+        if config and isinstance(config, Configuration):
+            return config[self.name]
+
+    def __repr__(self):
+        return '<%s [%s]>' % (self.__class__.__name__, self.name)
+
+
+class Option(object):
+    """Descriptor for configuration options."""
+
+    registry = {}
+
+    def accessor(self, section, name, default):
+        return section.get(name, default)
 
     @staticmethod
     def get_registry(compmgr=None):
         """Return the option registry, as a `dict` mapping `(section, key)`
         tuples to `Option` objects.
-        
+
         If `compmgr` is specified, only return options for components that are
         enabled in the given `ComponentManager`.
         """
-        if compmgr is None:
-            return Option.registry
-        
-        from trac.core import ComponentMeta
-        components = {}
-        for cls in ComponentMeta._components:
-            for attr in cls.__dict__.itervalues():
-                if isinstance(attr, Option):
-                    components[attr] = cls
-        
-        return dict(each for each in Option.registry.items()
-                    if each[1] not in components
-                       or compmgr.is_enabled(components[each[1]]))
-    
-    def __init__(self, section, name, default=None, doc=''):
+        return _get_registry(Option, compmgr)
+
+    def __init__(self, section, name, default=None, doc='',
+                 doc_domain='tracini'):
         """Create the configuration option.
-        
+
         @param section: the name of the configuration section this option
             belongs to
         @param name: the name of the option
@@ -550,7 +605,8 @@ class Option(object):
         self.name = name
         self.default = default
         self.registry[(self.section, self.name)] = self
-        self.__doc__ = doc
+        self.__doc__ = cleandoc(doc)
+        self.doc_domain = doc_domain
 
     def __get__(self, instance, owner):
         if instance is None:
@@ -560,29 +616,46 @@ class Option(object):
             section = config[self.section]
             value = self.accessor(section, self.name, self.default)
             return value
-        return None
 
     def __set__(self, instance, value):
-        raise AttributeError, 'can\'t set attribute'
+        raise AttributeError(_("Setting attribute is not allowed."))
 
     def __repr__(self):
         return '<%s [%s] "%s">' % (self.__class__.__name__, self.section,
                                    self.name)
 
+    def dumps(self, value):
+        """Return the value as a string to write to a trac.ini file"""
+        if value is None:
+            return ''
+        if value is True:
+            return 'enabled'
+        if value is False:
+            return 'disabled'
+        if isinstance(value, unicode):
+            return value
+        return to_unicode(value)
+
 
 class BoolOption(Option):
     """Descriptor for boolean configuration options."""
-    accessor = Section.getbool
+
+    def accessor(self, section, name, default):
+        return section.getbool(name, default)
 
 
 class IntOption(Option):
     """Descriptor for integer configuration options."""
-    accessor = Section.getint
+
+    def accessor(self, section, name, default):
+        return section.getint(name, default)
 
 
 class FloatOption(Option):
     """Descriptor for float configuration options."""
-    accessor = Section.getfloat
+
+    def accessor(self, section, name, default):
+        return section.getfloat(name, default)
 
 
 class ListOption(Option):
@@ -591,24 +664,30 @@ class ListOption(Option):
     """
 
     def __init__(self, section, name, default=None, sep=',', keep_empty=False,
-                 doc=''):
-        Option.__init__(self, section, name, default, doc)
+                 doc='', doc_domain='tracini'):
+        Option.__init__(self, section, name, default, doc, doc_domain)
         self.sep = sep
         self.keep_empty = keep_empty
 
     def accessor(self, section, name, default):
         return section.getlist(name, default, self.sep, self.keep_empty)
 
+    def dumps(self, value):
+        if isinstance(value, (list, tuple)):
+            return self.sep.join(Option.dumps(self, v) or '' for v in value)
+        return Option.dumps(self, value)
+
 
 class ChoiceOption(Option):
     """Descriptor for configuration options providing a choice among a list
     of items.
-    
+
     The default value is the first choice in the list.
     """
-    
-    def __init__(self, section, name, choices, doc=''):
-        Option.__init__(self, section, name, _to_utf8(choices[0]), doc)
+
+    def __init__(self, section, name, choices, doc='', doc_domain='tracini'):
+        Option.__init__(self, section, name, _to_utf8(choices[0]), doc,
+                        doc_domain)
         self.choices = set(_to_utf8(choice).strip() for choice in choices)
 
     def accessor(self, section, name, default):
@@ -621,17 +700,27 @@ class ChoiceOption(Option):
                       choices=', '.join('"%s"' % c
                                         for c in sorted(self.choices))))
         return value
-            
-    
+
+
 class PathOption(Option):
-    """Descriptor for file system path configuration options."""
-    accessor = Section.getpath
+    """Descriptor for file system path configuration options.
+
+    Relative paths are resolved to absolute paths using the directory
+    containing the configuration file as the reference.
+    """
+
+    def accessor(self, section, name, default):
+        return section.getpath(name, default)
 
 
 class ExtensionOption(Option):
+    """Name of a component implementing `interface`. Raises a
+    `ConfigurationError` if the component cannot be found in the list of
+    active components implementing the interface."""
 
-    def __init__(self, section, name, interface, default=None, doc=''):
-        Option.__init__(self, section, name, default, doc)
+    def __init__(self, section, name, interface, default=None, doc='',
+                 doc_domain='tracini'):
+        Option.__init__(self, section, name, default, doc, doc_domain)
         self.xtnpt = ExtensionPoint(interface)
 
     def __get__(self, instance, owner):
@@ -641,11 +730,14 @@ class ExtensionOption(Option):
         for impl in self.xtnpt.extensions(instance):
             if impl.__class__.__name__ == value:
                 return impl
-        raise AttributeError('Cannot find an implementation of the "%s" '
-                             'interface named "%s".  Please update the option '
-                             '%s.%s in trac.ini.'
-                             % (self.xtnpt.interface.__name__, value,
-                                self.section, self.name))
+        raise ConfigurationError(
+            tag_("Cannot find an implementation of the %(interface)s "
+                 "interface named %(implementation)s. Please check "
+                 "that the Component is enabled or update the option "
+                 "%(option)s in trac.ini.",
+                 interface=tag.tt(self.xtnpt.interface.__name__),
+                 implementation=tag.tt(value),
+                 option=tag.tt("[%s] %s" % (self.section, self.name))))
 
 
 class OrderedExtensionsOption(ListOption):
@@ -656,8 +748,9 @@ class OrderedExtensionsOption(ListOption):
     interface are returned, with those specified by the option ordered first."""
 
     def __init__(self, section, name, interface, default=None,
-                 include_missing=True, doc=''):
-        ListOption.__init__(self, section, name, default, doc=doc)
+                 include_missing=True, doc='', doc_domain='tracini'):
+        ListOption.__init__(self, section, name, default, doc=doc,
+                            doc_domain=doc_domain)
         self.xtnpt = ExtensionPoint(interface)
         self.include_missing = include_missing
 
@@ -666,9 +759,23 @@ class OrderedExtensionsOption(ListOption):
             return self
         order = ListOption.__get__(self, instance, owner)
         components = []
+        implementing_classes = []
         for impl in self.xtnpt.extensions(instance):
+            implementing_classes.append(impl.__class__.__name__)
             if self.include_missing or impl.__class__.__name__ in order:
                 components.append(impl)
+        not_found = sorted(set(order) - set(implementing_classes))
+        if not_found:
+            raise ConfigurationError(
+                tag_("Cannot find implementation(s) of the %(interface)s "
+                     "interface named %(implementation)s. Please check "
+                     "that the Component is enabled or update the option "
+                     "%(option)s in trac.ini.",
+                     interface=tag.tt(self.xtnpt.interface.__name__),
+                     implementation=tag(
+                         (', ' if idx != 0 else None, tag.tt(impl))
+                         for idx, impl in enumerate(not_found)),
+                     option=tag.tt("[%s] %s" % (self.section, self.name))))
 
         def compare(x, y):
             x, y = x.__class__.__name__, y.__class__.__name__
@@ -683,11 +790,11 @@ class OrderedExtensionsOption(ListOption):
 
 class ConfigurationAdmin(Component):
     """trac-admin command provider for trac.ini administration."""
-    
+
     implements(IAdminCommandProvider)
-    
+
     # IAdminCommandProvider methods
-    
+
     def get_admin_commands(self):
         yield ('config get', '<section> <option>',
                'Get the value of the given option in "trac.ini"',
