@@ -27,6 +27,8 @@ from trac.util.translation import _
 
 _like_escape_re = re.compile(r'([/_%])')
 
+_glob_escape_re = re.compile(r'[*?\[]')
+
 try:
     import pysqlite2.dbapi2 as sqlite
     have_pysqlite = 2
@@ -125,14 +127,14 @@ def _to_sql(table):
     sql.append(',\n'.join(coldefs) + '\n);')
     yield '\n'.join(sql)
     for index in table.indices:
-        unique = index.unique and 'UNIQUE' or ''
+        unique = 'UNIQUE' if index.unique else ''
         yield "CREATE %s INDEX %s_%s_idx ON %s (%s);" % (unique, table.name,
               '_'.join(index.columns), table.name, ','.join(index.columns))
 
 
 class SQLiteConnector(Component):
     """Database connector for SQLite.
-    
+
     Database URLs should be of the form:
     {{{
     sqlite:path/to/trac.db
@@ -140,10 +142,12 @@ class SQLiteConnector(Component):
     """
     implements(IDatabaseConnector)
 
-    extensions = ListOption('sqlite', 'extensions', 
+    extensions = ListOption('sqlite', 'extensions',
         doc="""Paths to sqlite extensions, relative to Trac environment's
         directory or absolute. (''since 0.12'')""")
 
+    memory_cnx = None
+
     def __init__(self):
         self._version = None
         self.error = None
@@ -159,7 +163,7 @@ class SQLiteConnector(Component):
         elif (2, 5, 2) <= sqlite.version_info < (2, 5, 5):
             self.error = _("PySqlite 2.5.2 - 2.5.4 break Trac, please use "
                            "2.5.5 or higher")
-        yield ('sqlite', self.error and -1 or 1)
+        yield ('sqlite', -1 if self.error else 1)
 
     def get_connection(self, path, log=None, params={}):
         if not self._version:
@@ -176,22 +180,35 @@ class SQLiteConnector(Component):
                     extpath = os.path.join(self.env.path, extpath)
                 self._extensions.append(extpath)
         params['extensions'] = self._extensions
-        return SQLiteConnection(path, log, params)
+        if path == ':memory:':
+            if not self.memory_cnx:
+                self.memory_cnx = SQLiteConnection(path, log, params)
+            return self.memory_cnx
+        else:
+            return SQLiteConnection(path, log, params)
 
-    def init_db(self, path, log=None, params={}):
+    def get_exceptions(self):
+        return sqlite
+
+    def init_db(self, path, schema=None, log=None, params={}):
         if path != ':memory:':
             # make the directory to hold the database
             if os.path.exists(path):
-                raise TracError(_('Database already exists at %(path)s',
+                raise TracError(_("Database already exists at %(path)s",
                                   path=path))
             dir = os.path.dirname(path)
             if not os.path.exists(dir):
                 os.makedirs(dir)
-        if isinstance(path, unicode): # needed with 2.4.0
-            path = path.encode('utf-8')
-        cnx = sqlite.connect(path, timeout=int(params.get('timeout', 10000)))
+            if isinstance(path, unicode): # needed with 2.4.0
+                path = path.encode('utf-8')
+            # this direct connect will create the database if needed
+            cnx = sqlite.connect(path,
+                                 timeout=int(params.get('timeout', 10000)))
+        else:
+            cnx = self.get_connection(path, log, params)
         cursor = cnx.cursor()
-        from trac.db_default import schema
+        if schema is None:
+            from trac.db_default import schema
         for table in schema:
             for stmt in self.to_sql(table):
                 cursor.execute(stmt)
@@ -203,7 +220,7 @@ class SQLiteConnector(Component):
     def alter_column_types(self, table, columns):
         """Yield SQL statements altering the type of one or more columns of
         a table.
-        
+
         Type changes are specified as a `columns` dict mapping column names
         to `(from, to)` SQL type tuples.
         """
@@ -240,7 +257,8 @@ class SQLiteConnection(ConnectionWrapper):
                              and sqlite.version_info >= (2, 5, 0)
 
     def __init__(self, path, log=None, params={}):
-        assert have_pysqlite > 0
+        if have_pysqlite == 0:
+            raise TracError(_("Cannot load Python bindings for SQLite"))
         self.cnx = None
         if path != ':memory:':
             if not os.access(path, os.F_OK):
@@ -271,7 +289,7 @@ class SQLiteConnection(ConnectionWrapper):
             for ext in extensions:
                 cnx.load_extension(ext)
             cnx.enable_load_extension(False)
-       
+
         ConnectionWrapper.__init__(self, cnx, log)
 
     def cursor(self):
@@ -297,6 +315,24 @@ class SQLiteConnection(ConnectionWrapper):
     def concat(self, *args):
         return '||'.join(args)
 
+    def drop_table(self, table):
+        cursor = self.cursor()
+        cursor.execute("DROP TABLE IF EXISTS " + self.quote(table))
+
+    def get_column_names(self, table):
+        cursor = self.cnx.cursor()
+        rows = cursor.execute("PRAGMA table_info(%s)"
+                              % self.quote(table))
+        return [row[1] for row in rows]
+
+    def get_last_id(self, cursor, table, column='id'):
+        return cursor.lastrowid
+
+    def get_table_names(self):
+        rows = self.execute("""
+            SELECT name FROM sqlite_master WHERE type='table'""")
+        return [row[0] for row in rows]
+
     def like(self):
         """Return a case-insensitive LIKE clause."""
         if sqlite_version >= (3, 1, 0):
@@ -310,13 +346,18 @@ class SQLiteConnection(ConnectionWrapper):
         else:
             return text
 
+    def prefix_match(self):
+        """Return a case sensitive prefix-matching operator."""
+        return 'GLOB %s'
+
+    def prefix_match_value(self, prefix):
+        """Return a value for case sensitive prefix-matching operator."""
+        return _glob_escape_re.sub(lambda m: '[%s]' % m.group(0), prefix) + '*'
+
     def quote(self, identifier):
         """Return the quoted identifier."""
-        return "`%s`" % identifier
+        return "`%s`" % identifier.replace('`', '``')
 
-    def get_last_id(self, cursor, table, column='id'):
-        return cursor.lastrowid
-    
     def update_sequence(self, cursor, table, column='id'):
         # SQLite handles sequence updates automagically
         # http://www.sqlite.org/autoinc.html
