@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2009 Edgewall Software
+# Copyright (C) 2003-2014 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
@@ -21,8 +21,8 @@ from __future__ import with_statement
 
 import doctest
 import os
-import unittest
 import sys
+import unittest
 
 try:
     from babel import Locale
@@ -30,14 +30,13 @@ try:
 except ImportError:
     locale_en = None
 
-from trac.config import Configuration
-from trac.core import Component, ComponentManager
-from trac.env import Environment
-from trac.db.api import _parse_db_str, DatabaseManager
-from trac.db.sqlite_backend import SQLiteConnection
-from trac.db.util import ConnectionWrapper
-import trac.db.postgres_backend
 import trac.db.mysql_backend
+import trac.db.postgres_backend
+import trac.db.sqlite_backend
+from trac.config import Configuration
+from trac.core import ComponentManager
+from trac.db.api import DatabaseManager, _parse_db_str
+from trac.env import Environment
 from trac.ticket.default_workflow import load_workflow_config_snippet
 from trac.util import translation
 
@@ -144,7 +143,8 @@ class TestSetup(unittest.TestSuite):
         return result
 
     def _wrapped_run(self, *args, **kwargs):
-        "Python 2.7 / unittest2 compatibility - there must be a better way..."
+        """Python 2.7 / unittest2 compatibility - there must be a better
+        way..."""
         self.setUp()
         if hasattr(self, 'fixture'):
             for test in self._tests:
@@ -153,6 +153,7 @@ class TestSetup(unittest.TestSuite):
         unittest.TestSuite._wrapped_run(self, *args, **kwargs)
         self.tearDown()
 
+
 class TestCaseSetup(unittest.TestCase):
     def setFixture(self, fixture):
         self.fixture = fixture
@@ -164,7 +165,7 @@ def get_dburi():
     dburi = os.environ.get('TRAC_TEST_DB_URI')
     if dburi:
         scheme, db_prop = _parse_db_str(dburi)
-        # Assume the schema 'tractest' for Postgres
+        # Assume the schema 'tractest' for PostgreSQL
         if scheme == 'postgres' and \
                 not db_prop.get('params', {}).get('schema'):
             if '?' in dburi:
@@ -176,9 +177,8 @@ def get_dburi():
 
 
 def reset_sqlite_db(env, db_prop):
-    dbname = os.path.basename(db_prop['path'])
     with env.db_transaction as db:
-        tables = db("SELECT name FROM sqlite_master WHERE type='table'")
+        tables = db.get_table_names()
         for table in tables:
             db("DELETE FROM %s" % table)
         return tables
@@ -189,24 +189,24 @@ def reset_postgres_db(env, db_prop):
         dbname = db.schema
         if dbname:
             # reset sequences
-            # information_schema.sequences view is available in PostgreSQL 8.2+
-            # however Trac supports PostgreSQL 8.0+, uses
+            # information_schema.sequences view is available in
+            # PostgreSQL 8.2+ however Trac supports PostgreSQL 8.0+, uses
             # pg_get_serial_sequence()
-            for seq in db("""
-                    SELECT sequence_name FROM (
-                        SELECT pg_get_serial_sequence(%s||table_name,
-                                                      column_name)
-                               AS sequence_name
-                        FROM information_schema.columns
-                        WHERE table_schema=%s) AS tab
-                    WHERE sequence_name IS NOT NULL""",
-                    (dbname + '.', dbname)):
+            seqs = [seq for seq, in db("""
+                SELECT sequence_name
+                FROM (
+                    SELECT pg_get_serial_sequence(
+                        quote_ident(table_schema) || '.' ||
+                        quote_ident(table_name), column_name) AS sequence_name
+                    FROM information_schema.columns
+                    WHERE table_schema=%s) AS tab
+                WHERE sequence_name IS NOT NULL""", (dbname,))]
+            for seq in seqs:
                 db("ALTER SEQUENCE %s RESTART WITH 1" % seq)
             # clear tables
-            tables = db("""SELECT table_name FROM information_schema.tables
-                           WHERE table_schema=%s""", (dbname,))
+            tables = db.get_table_names()
             for table in tables:
-                db("DELETE FROM %s" % table)
+                db("DELETE FROM %s" % db.quote(table))
             # PostgreSQL supports TRUNCATE TABLE as well
             # (see http://www.postgresql.org/docs/8.1/static/sql-truncate.html)
             # but on the small tables used here, DELETE is actually much faster
@@ -217,12 +217,18 @@ def reset_mysql_db(env, db_prop):
     dbname = os.path.basename(db_prop['path'])
     if dbname:
         with env.db_transaction as db:
-            tables = db("""SELECT table_name FROM information_schema.tables
+            tables = db("""SELECT table_name, auto_increment
+                           FROM information_schema.tables
                            WHERE table_schema=%s""", (dbname,))
-            for table in tables:
-                # TRUNCATE TABLE is prefered to DELETE FROM, as we need to reset
-                # the auto_increment in MySQL.
-                db("TRUNCATE TABLE %s" % table)
+            for table, auto_increment in tables:
+                if auto_increment is None or auto_increment == 1:
+                    # DELETE FROM is preferred to TRUNCATE TABLE, as the
+                    # auto_increment is not used or it is 1.
+                    db("DELETE FROM %s" % table)
+                else:
+                    # TRUNCATE TABLE is preferred to DELETE FROM, as we
+                    # need to reset the auto_increment in MySQL.
+                    db("TRUNCATE TABLE %s" % table)
             return tables
 
 
@@ -231,8 +237,8 @@ def reset_mysql_db(env, db_prop):
 class EnvironmentStub(Environment):
     """A stub of the trac.env.Environment object for testing."""
 
-    href = abs_href = None
     global_databasemanager = None
+    required = False
 
     def __init__(self, default_data=False, enable=None, disable=None,
                  path=None, destroying=False):
@@ -242,9 +248,21 @@ class EnvironmentStub(Environment):
                              defaults.
         :param enable: A list of component classes or name globs to
                        activate in the stub environment.
+        :param disable: A list of component classes or name globs to
+                        deactivate in the stub environment.
+        :param path: The location of the environment in the file system.
+                     No files or directories are created when specifying
+                     this parameter.
+        :param destroying: If True, the database will not be reset. This is
+                           useful for cases when the object is being
+                           constructed in order to call `destroy_db`.
         """
+        if enable is not None and not isinstance(enable, (list, tuple)):
+            raise TypeError('Keyword argument "enable" must be a list')
+        if disable is not None and not isinstance(disable, (list, tuple)):
+            raise TypeError('Keyword argument "disable" must be a list')
+
         ComponentManager.__init__(self)
-        Component.__init__(self)
 
         self.systeminfo = []
 
@@ -266,7 +284,7 @@ class EnvironmentStub(Environment):
         if enable is not None:
             self.config.set('components', 'trac.*', 'disabled')
         else:
-            self.config.set('components', 'tracopt.versioncontrol.svn.*',
+            self.config.set('components', 'tracopt.versioncontrol.*',
                             'enabled')
         for name_or_class in enable or ():
             config_key = self._component_name(name_or_class)
@@ -285,21 +303,17 @@ class EnvironmentStub(Environment):
 
         init_global = False
         if self.global_databasemanager:
-            self.components[DatabaseManager] = global_databasemanager
+            self.components[DatabaseManager] = self.global_databasemanager
         else:
             self.config.set('trac', 'database', self.dburi)
             self.global_databasemanager = DatabaseManager(self)
             self.config.set('trac', 'debug_sql', True)
-            self.config.set('logging', 'log_type', 'stderr')
-            self.config.set('logging', 'log_level', 'DEBUG')
             init_global = not destroying
 
         if default_data or init_global:
             self.reset_db(default_data)
 
-        from trac.web.href import Href
-        self.href = Href('/trac.cgi')
-        self.abs_href = Href('http://example.org/trac.cgi')
+        self.config.set('trac', 'base_url', 'http://example.org/trac.cgi')
 
         self.known_users = []
         translation.activate(locale_en)
@@ -315,28 +329,24 @@ class EnvironmentStub(Environment):
         remove_sqlite_db = False
         try:
             with self.db_transaction as db:
-                db.rollback() # make sure there's no transaction in progress
+                db.rollback()  # make sure there's no transaction in progress
                 # check the database version
-                database_version = db(
-                    "SELECT value FROM system WHERE name='database_version'")
-                if database_version:
-                    database_version = int(database_version[0][0])
-                if database_version == db_default.db_version:
-                    # same version, simply clear the tables (faster)
-                    m = sys.modules[__name__]
-                    reset_fn = 'reset_%s_db' % scheme
-                    if hasattr(m, reset_fn):
-                        tables = getattr(m, reset_fn)(self, db_prop)
-                else:
-                    # different version or version unknown, drop the tables
-                    remove_sqlite_db = True
-                    self.destroy_db(scheme, db_prop)
-        except Exception, e:
+                database_version = self.database_version
+        except Exception:
             # "Database not found ...",
             # "OperationalError: no such table: system" or the like
             pass
-
-        db = None # as we might shutdown the pool     FIXME no longer needed!
+        else:
+            if database_version == db_default.db_version:
+                # same version, simply clear the tables (faster)
+                m = sys.modules[__name__]
+                reset_fn = 'reset_%s_db' % scheme
+                if hasattr(m, reset_fn):
+                    tables = getattr(m, reset_fn)(self, db_prop)
+            else:
+                # different version or version unknown, drop the tables
+                remove_sqlite_db = True
+                self.destroy_db(scheme, db_prop)
 
         if scheme == 'sqlite' and remove_sqlite_db:
             path = db_prop['path']
@@ -354,12 +364,14 @@ class EnvironmentStub(Environment):
                 self.global_databasemanager.shutdown()
 
         with self.db_transaction as db:
+            if scheme == 'sqlite':
+                # Speed-up tests with SQLite database
+                db("PRAGMA synchronous = OFF")
             if default_data:
                 for table, cols, vals in db_default.get_data(db):
                     db.executemany("INSERT INTO %s (%s) VALUES (%s)"
                                    % (table, ','.join(cols),
-                                      ','.join(['%s' for c in cols])),
-                                   vals)
+                                      ','.join(['%s'] * len(cols))), vals)
             else:
                 db("INSERT INTO system (name, value) VALUES (%s, %s)",
                    ('database_version', str(db_default.db_version)))
@@ -370,12 +382,9 @@ class EnvironmentStub(Environment):
         try:
             with self.db_transaction as db:
                 if scheme == 'postgres' and db.schema:
-                    db('DROP SCHEMA "%s" CASCADE' % db.schema)
+                    db('DROP SCHEMA %s CASCADE' % db.quote(db.schema))
                 elif scheme == 'mysql':
-                    dbname = os.path.basename(db_prop['path'])
-                    for table in db("""
-                          SELECT table_name FROM information_schema.tables
-                          WHERE table_schema=%s""", (dbname,)):
+                    for table in db.get_table_names():
                         db("DROP TABLE IF EXISTS `%s`" % table)
         except Exception:
             # "TracError: Database not found...",
@@ -383,7 +392,7 @@ class EnvironmentStub(Environment):
             pass
         return False
 
-    # overriden
+    # overridden
 
     def is_component_enabled(self, cls):
         if self._component_name(cls).startswith('__main__.'):
@@ -410,11 +419,13 @@ def locate(fn):
 
 INCLUDE_FUNCTIONAL_TESTS = True
 
+
 def suite():
     import trac.tests
     import trac.admin.tests
     import trac.db.tests
     import trac.mimeview.tests
+    import trac.timeline.tests
     import trac.ticket.tests
     import trac.util.tests
     import trac.versioncontrol.tests
@@ -423,17 +434,17 @@ def suite():
     import trac.wiki.tests
     import tracopt.mimeview.tests
     import tracopt.perm.tests
+    import tracopt.ticket.tests
     import tracopt.versioncontrol.git.tests
     import tracopt.versioncontrol.svn.tests
 
     suite = unittest.TestSuite()
     suite.addTest(trac.tests.basicSuite())
-    if INCLUDE_FUNCTIONAL_TESTS:
-        suite.addTest(trac.tests.functionalSuite())
     suite.addTest(trac.admin.tests.suite())
     suite.addTest(trac.db.tests.suite())
     suite.addTest(trac.mimeview.tests.suite())
     suite.addTest(trac.ticket.tests.suite())
+    suite.addTest(trac.timeline.tests.suite())
     suite.addTest(trac.util.tests.suite())
     suite.addTest(trac.versioncontrol.tests.suite())
     suite.addTest(trac.versioncontrol.web_ui.tests.suite())
@@ -441,14 +452,17 @@ def suite():
     suite.addTest(trac.wiki.tests.suite())
     suite.addTest(tracopt.mimeview.tests.suite())
     suite.addTest(tracopt.perm.tests.suite())
+    suite.addTest(tracopt.ticket.tests.suite())
     suite.addTest(tracopt.versioncontrol.git.tests.suite())
     suite.addTest(tracopt.versioncontrol.svn.tests.suite())
     suite.addTest(doctest.DocTestSuite(sys.modules[__name__]))
-
+    if INCLUDE_FUNCTIONAL_TESTS:
+        suite.addTest(trac.tests.functionalSuite())
     return suite
 
+
 if __name__ == '__main__':
-    #FIXME: this is a bit inelegant
+    # FIXME: this is a bit inelegant
     if '--skip-functional-tests' in sys.argv:
         sys.argv.remove('--skip-functional-tests')
         INCLUDE_FUNCTIONAL_TESTS = False
