@@ -24,7 +24,7 @@ from trac.config import *
 from trac.core import *
 from trac.perm import IPermissionRequestor, PermissionCache, PermissionSystem
 from trac.resource import IResourceManager
-from trac.util import Ranges
+from trac.util import Ranges, as_int
 from trac.util.text import shorten_line
 from trac.util.translation import _, N_, gettext
 from trac.wiki import IWikiSyntaxProvider, WikiParser
@@ -40,19 +40,20 @@ class ITicketActionController(Interface):
 
     def get_ticket_actions(req, ticket):
         """Return an iterable of `(weight, action)` tuples corresponding to
-        the actions that are contributed by this component.
-        That list may vary given the current state of the ticket and the
-        actual request parameter.
+        the actions that are contributed by this component. The list is
+        dependent on the current state of the ticket and the actual request
+        parameter.
 
         `action` is a key used to identify that particular action.
         (note that 'history' and 'diff' are reserved and should not be used
         by plugins)
-        
+
         The actions will be presented on the page in descending order of the
         integer weight. The first action in the list is used as the default
         action.
 
-        When in doubt, use a weight of 0."""
+        When in doubt, use a weight of 0.
+        """
 
     def get_all_status():
         """Returns an iterable of all the possible values for the ''status''
@@ -69,7 +70,7 @@ class ITicketActionController(Interface):
         `label` is a short text that will be used when listing the action,
         `control` is the markup for the action control and `hint` should
         explain what will happen if this action is taken.
-        
+
         This method will only be called if the controller claimed to handle
         the given `action` in the call to `get_ticket_actions`.
 
@@ -114,7 +115,7 @@ class ITicketChangeListener(Interface):
 
     def ticket_changed(ticket, comment, author, old_values):
         """Called when a ticket is modified.
-        
+
         `old_values` is a dictionary containing the previous values of the
         fields that have changed.
         """
@@ -122,6 +123,15 @@ class ITicketChangeListener(Interface):
     def ticket_deleted(ticket):
         """Called when a ticket is deleted."""
 
+    def ticket_comment_modified(ticket, cdate, author, comment, old_comment):
+        """Called when a ticket comment is modified."""
+
+    def ticket_change_deleted(ticket, cdate, changes):
+        """Called when a ticket change is deleted.
+
+        `changes` is a dictionary of tuple `(oldvalue, newvalue)`
+        containing the ticket change of the fields that have changed."""
+
 
 class ITicketManipulator(Interface):
     """Miscellaneous manipulation of ticket workflow features."""
@@ -132,7 +142,7 @@ class ITicketManipulator(Interface):
 
     def validate_ticket(req, ticket):
         """Validate a ticket after it's been populated from user input.
-        
+
         Must return a list of `(field, message)` tuples, one for each problem
         detected. `field` can be `None` to indicate an overall problem with the
         ticket. Therefore, a return value of `[]` means everything is OK."""
@@ -162,7 +172,11 @@ class TicketSystem(Component):
 
     change_listeners = ExtensionPoint(ITicketChangeListener)
     milestone_change_listeners = ExtensionPoint(IMilestoneChangeListener)
-    
+
+    ticket_custom_section = ConfigSection('ticket-custom',
+        """In this section, you can define additional fields for tickets. See
+        TracTicketsCustomFields for more details.""")
+
     action_controllers = OrderedExtensionsOption('ticket', 'workflow',
         ITicketActionController, default='ConfigurableTicketWorkflow',
         include_missing=False,
@@ -173,8 +187,12 @@ class TicketSystem(Component):
         """Make the owner field of tickets use a drop-down menu.
         Be sure to understand the performance implications before activating
         this option. See
-        [TracTickets#Assign-toasDrop-DownList Assign-to as Drop-Down List]
-        (''since 0.9'').""")
+        [TracTickets#Assign-toasDrop-DownList Assign-to as Drop-Down List].
+
+        Please note that e-mail addresses are '''not''' obfuscated in the
+        resulting drop-down menu, so this option should not be used if
+        e-mail addresses must remain protected.
+        (''since 0.9'')""")
 
     default_version = Option('ticket', 'default_version', '',
         """Default version for newly created tickets.""")
@@ -203,7 +221,7 @@ class TicketSystem(Component):
     default_keywords = Option('ticket', 'default_keywords', '',
         """Default keywords for newly created tickets.""")
 
-    default_owner = Option('ticket', 'default_owner', '',
+    default_owner = Option('ticket', 'default_owner', '< default >',
         """Default owner for newly created tickets.""")
 
     default_cc = Option('ticket', 'default_cc', '',
@@ -214,8 +232,8 @@ class TicketSystem(Component):
         (''since 0.11'').""")
 
     def __init__(self):
-        self.log.debug('action controllers for ticket workflow: %r' % 
-                [c.__class__.__name__ for c in self.action_controllers])
+        self.log.debug('action controllers for ticket workflow: %r',
+                       [c.__class__.__name__ for c in self.action_controllers])
 
     # Public API
 
@@ -244,8 +262,10 @@ class TicketSystem(Component):
 
     def get_ticket_field_labels(self):
         """Produce a (name,label) mapping from `get_ticket_fields`."""
-        return dict((f['name'], f['label']) for f in
-                    TicketSystem(self.env).get_ticket_fields())
+        labels = dict((f['name'], f['label'])
+                      for f in self.get_ticket_fields())
+        labels['attachment'] = _("Attachment")
+        return labels
 
     def get_ticket_fields(self):
         """Returns list of fields available for tickets.
@@ -278,11 +298,10 @@ class TicketSystem(Component):
         fields.append({'name': 'reporter', 'type': 'text',
                        'label': N_('Reporter')})
 
-        # Owner field, by default text but can be changed dynamically 
+        # Owner field, by default text but can be changed dynamically
         # into a drop-down depending on configuration (restrict_owner=true)
-        field = {'name': 'owner', 'label': N_('Owner')}
-        field['type'] = 'text'
-        fields.append(field)
+        fields.append({'name': 'owner', 'type': 'text',
+                       'label': N_('Owner')})
 
         # Description
         fields.append({'name': 'description', 'type': 'textarea',
@@ -314,9 +333,10 @@ class TicketSystem(Component):
             fields.append(field)
 
         # Advanced text fields
-        fields.append({'name': 'keywords', 'type': 'text',
+        fields.append({'name': 'keywords', 'type': 'text', 'format': 'list',
                        'label': N_('Keywords')})
-        fields.append({'name': 'cc', 'type': 'text', 'label': N_('Cc')})
+        fields.append({'name': 'cc', 'type': 'text',  'format': 'list',
+                       'label': N_('Cc')})
 
         # Date/time fields
         fields.append({'name': 'time', 'type': 'time',
@@ -353,7 +373,7 @@ class TicketSystem(Component):
     def custom_fields(self, db):
         """Return the list of custom ticket fields available for tickets."""
         fields = []
-        config = self.config['ticket-custom']
+        config = self.ticket_custom_section
         for name in [option for option, value in config.options()
                      if '.' not in option]:
             field = {
@@ -376,7 +396,8 @@ class TicketSystem(Component):
                 field['height'] = config.getint(name + '.rows')
             fields.append(field)
 
-        fields.sort(lambda x, y: cmp(x['order'], y['order']))
+        fields.sort(lambda x, y: cmp((x['order'], x['name']),
+                                     (y['order'], y['name'])))
         return fields
 
     def get_field_synonyms(self):
@@ -391,28 +412,44 @@ class TicketSystem(Component):
         """
         if self.restrict_owner:
             field['type'] = 'select'
-            possible_owners = []
+            allowed_owners = self.get_allowed_owners(ticket)
+            allowed_owners.insert(0, '< default >')
+            field['options'] = allowed_owners
+            field['optional'] = True
+
+    def get_allowed_owners(self, ticket=None):
+        """Returns a list of permitted ticket owners (those possessing the
+        TICKET_MODIFY permission). Returns `None` if the option `[ticket]`
+        `restrict_owner` is `False`.
+
+        If `ticket` is not `None`, fine-grained permission checks are used
+        to determine the allowed owners for the specified resource.
+
+        :since: 1.0.3
+        """
+        if self.restrict_owner:
+            allowed_owners = []
             for user in PermissionSystem(self.env) \
-                    .get_users_with_permission('TICKET_MODIFY'):
+                        .get_users_with_permission('TICKET_MODIFY'):
                 if not ticket or \
                         'TICKET_MODIFY' in PermissionCache(self.env, user,
                                                            ticket.resource):
-                    possible_owners.append(user)
-            possible_owners.sort()
-            field['options'] = possible_owners
-            field['optional'] = True
+                    allowed_owners.append(user)
+            allowed_owners.sort()
+            return allowed_owners
 
     # IPermissionRequestor methods
 
     def get_permission_actions(self):
         return ['TICKET_APPEND', 'TICKET_CREATE', 'TICKET_CHGPROP',
                 'TICKET_VIEW', 'TICKET_EDIT_CC', 'TICKET_EDIT_DESCRIPTION',
-                'TICKET_EDIT_COMMENT',
+                'TICKET_EDIT_COMMENT', 'TICKET_BATCH_MODIFY',
                 ('TICKET_MODIFY', ['TICKET_APPEND', 'TICKET_CHGPROP']),
                 ('TICKET_ADMIN', ['TICKET_CREATE', 'TICKET_MODIFY',
                                   'TICKET_VIEW', 'TICKET_EDIT_CC',
                                   'TICKET_EDIT_DESCRIPTION',
-                                  'TICKET_EDIT_COMMENT'])]
+                                  'TICKET_EDIT_COMMENT',
+                                  'TICKET_BATCH_MODIFY'])]
 
     # IWikiSyntaxProvider methods
 
@@ -444,22 +481,26 @@ class TicketSystem(Component):
                 from trac.ticket.model import Ticket
                 if Ticket.id_is_valid(num) and \
                         'TICKET_VIEW' in formatter.perm(ticket):
-                    # TODO: watch #6436 and when done, attempt to retrieve 
-                    #       ticket directly (try: Ticket(self.env, num) ...)
-                    cursor = formatter.db.cursor() 
-                    cursor.execute("SELECT type,summary,status,resolution "
-                                   "FROM ticket WHERE id=%s", (str(num),)) 
-                    for type, summary, status, resolution in cursor:
+                    # TODO: attempt to retrieve ticket view directly,
+                    #       something like: t = Ticket.view(num)
+                    for type, summary, status, resolution in \
+                            self.env.db_query("""
+                            SELECT type, summary, status, resolution
+                            FROM ticket WHERE id=%s
+                            """, (str(num),)):
                         title = self.format_summary(summary, status,
                                                     resolution, type)
                         href = formatter.href.ticket(num) + params + fragment
-                        return tag.a(label, class_='%s ticket' % status, 
-                                     title=title, href=href)
+                        return tag.a(label, title=title, href=href,
+                                     class_='%s ticket' % status)
             else:
                 ranges = str(r)
                 if params:
                     params = '&' + params[1:]
-                return tag.a(label, title='Tickets '+ranges,
+                label_wrap = label.replace(',', u',\u200b')
+                ranges_wrap = ranges.replace(',', u', ')
+                return tag.a(label_wrap,
+                             title=_("Tickets %(ranges)s", ranges=ranges_wrap),
                              href=formatter.href.query(id=ranges) + params)
         except ValueError:
             pass
@@ -473,19 +514,46 @@ class TicketSystem(Component):
                 cnum, realm, id = elts
                 if cnum != 'description' and cnum and not cnum[0].isdigit():
                     realm, id, cnum = elts # support old comment: style
+                id = as_int(id, None)
                 resource = formatter.resource(realm, id)
         else:
             resource = formatter.resource
             cnum = target
 
-        if resource:
-            href = "%s#comment:%s" % (formatter.href.ticket(resource.id), cnum)
-            title = _("Comment %(cnum)s for Ticket #%(id)s", cnum=cnum,
-                      id=resource.id)
-            return tag.a(label, href=href, title=title)
-        else:
-            return label
- 
+        if resource and resource.id and resource.realm == 'ticket' and \
+                cnum and (all(c.isdigit() for c in cnum) or cnum == 'description'):
+            href = title = class_ = None
+            if self.resource_exists(resource):
+                from trac.ticket.model import Ticket
+                ticket = Ticket(self.env, resource.id)
+                if cnum != 'description' and not ticket.get_change(cnum):
+                    title = _("ticket comment does not exist")
+                    class_ = 'missing ticket'
+                elif 'TICKET_VIEW' in formatter.perm(resource):
+                    href = formatter.href.ticket(resource.id) + \
+                           "#comment:%s" % cnum
+                    if resource.id != formatter.resource.id:
+                        if cnum == 'description':
+                            title = _("Description for Ticket #%(id)s",
+                                      id=resource.id)
+                        else:
+                            title = _("Comment %(cnum)s for Ticket #%(id)s",
+                                      cnum=cnum, id=resource.id)
+                        class_ = ticket['status'] + ' ticket'
+                    else:
+                        title = _("Description") if cnum == 'description' \
+                                                 else _("Comment %(cnum)s",
+                                                        cnum=cnum)
+                        class_ = 'ticket'
+                else:
+                    title = _("no permission to view ticket")
+                    class_ = 'forbidden ticket'
+            else:
+                title = _("ticket does not exist")
+                class_ = 'missing ticket'
+            return tag.a(label, class_=class_, href=href, title=title)
+        return label
+
     # IResourceManager methods
 
     def get_resource_realms(self):
@@ -530,16 +598,13 @@ class TicketSystem(Component):
         >>> resource_exists(env, t.resource)
         True
         """
-        db = self.env.get_read_db()
-        cursor = db.cursor()
-        cursor.execute("SELECT id FROM ticket WHERE id=%s", (resource.id,))
-        latest_exists = bool(cursor.fetchall())
-        if latest_exists:
+        if self.env.db_query("SELECT id FROM ticket WHERE id=%s",
+                             (resource.id,)):
             if resource.version is None:
                 return True
-            cursor.execute("""
-                SELECT count(distinct time) FROM ticket_change WHERE ticket=%s
+            revcount = self.env.db_query("""
+                SELECT count(DISTINCT time) FROM ticket_change WHERE ticket=%s
                 """, (resource.id,))
-            return cursor.fetchone()[0] >= resource.version
+            return revcount[0][0] >= resource.version
         else:
             return False
