@@ -3,7 +3,7 @@
 # Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2004-2005 Christopher Lenz <cmlenz@gmx.de>
-# Copyright (C) 2005-2006 Christian Boos <cboos@neuf.fr>
+# Copyright (C) 2005-2006 Christian Boos <cboos@edgewall.org>
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -25,19 +25,21 @@ from genshi.builder import tag
 
 from trac.config import IntOption, BoolOption
 from trac.core import *
-from trac.mimeview import Context
 from trac.perm import IPermissionRequestor
 from trac.timeline.api import ITimelineEventProvider
 from trac.util import as_int
-from trac.util.datefmt import format_date, format_datetime, parse_date, \
-                              to_utimestamp, utc, pretty_timedelta
+from trac.util.datefmt import format_date, format_datetime, format_time, \
+                              parse_date, to_utimestamp, to_datetime, utc, \
+                              pretty_timedelta, user_time
 from trac.util.text import exception_to_unicode, to_unicode
 from trac.util.translation import _, tag_
 from trac.web import IRequestHandler, IRequestFilter
-from trac.web.chrome import add_link, add_stylesheet, prevnext_nav, Chrome, \
-                            INavigationContributor, ITemplateProvider
-                            
+from trac.web.chrome import (Chrome, INavigationContributor, ITemplateProvider,
+                             add_link, add_stylesheet, auth_link, prevnext_nav,
+                             web_context)
 from trac.wiki.api import IWikiSyntaxProvider
+from trac.wiki.formatter import concat_path_query_fragment, \
+                                split_url_into_path_query_fragment
 
 
 class TimelineModule(Component):
@@ -52,7 +54,7 @@ class TimelineModule(Component):
         (''since 0.9.'')""")
 
     max_daysback = IntOption('timeline', 'max_daysback', 90,
-        """Maximum number of days (-1 for unlimited) displayable in the 
+        """Maximum number of days (-1 for unlimited) displayable in the
         Timeline. (''since 0.11'')""")
 
     abbreviated_messages = BoolOption('timeline', 'abbreviated_messages',
@@ -73,7 +75,7 @@ class TimelineModule(Component):
     def get_navigation_items(self, req):
         if 'TIMELINE_VIEW' in req.perm:
             yield ('mainnav', 'timeline',
-                   tag.a(_('Timeline'), href=req.href.timeline(), accesskey=2))
+                   tag.a(_("Timeline"), href=req.href.timeline(), accesskey=2))
 
     # IPermissionRequestor methods
 
@@ -89,18 +91,28 @@ class TimelineModule(Component):
         req.perm.assert_permission('TIMELINE_VIEW')
 
         format = req.args.get('format')
-        maxrows = int(req.args.get('max', format == 'rss' and 50 or 0))
+        maxrows = int(req.args.get('max', 50 if format == 'rss' else 0))
+        lastvisit = int(req.session.get('timeline.lastvisit', '0'))
+
+        # indication of new events is unchanged when form is updated by user
+        revisit = any(a in req.args for a in ['update', 'from', 'daysback',
+                                              'author'])
+        if revisit:
+            lastvisit = int(req.session.get('timeline.nextlastvisit',
+                                            lastvisit))
 
         # Parse the from date and adjust the timestamp to the last second of
         # the day
         fromdate = today = datetime.now(req.tz)
+        yesterday = to_datetime(today.replace(tzinfo=None) - timedelta(days=1),
+                                req.tz)
         precisedate = precision = None
         if 'from' in req.args:
             # Acquire from date only from non-blank input
             reqfromdate = req.args['from'].strip()
             if reqfromdate:
-                precisedate = parse_date(reqfromdate, req.tz)
-                fromdate = precisedate
+                precisedate = user_time(req, parse_date, reqfromdate)
+                fromdate = precisedate.astimezone(req.tz)
             precision = req.args.get('precision', '')
             if precision.startswith('second'):
                 precision = timedelta(seconds=1)
@@ -110,11 +122,12 @@ class TimelineModule(Component):
                 precision = timedelta(hours=1)
             else:
                 precision = None
-        fromdate = fromdate.replace(hour=23, minute=59, second=59,
-                                    microsecond=999999)
+        fromdate = to_datetime(datetime(fromdate.year, fromdate.month,
+                                        fromdate.day, 23, 59, 59, 999999),
+                               req.tz)
 
         daysback = as_int(req.args.get('daysback'),
-                          format == 'rss' and 90 or None)
+                          90 if format == 'rss' else None)
         if daysback is None:
             daysback = as_int(req.session.get('timeline.daysback'), None)
         if daysback is None:
@@ -130,12 +143,12 @@ class TimelineModule(Component):
 
         data = {'fromdate': fromdate, 'daysback': daysback,
                 'authors': authors,
-                'today': format_date(today, tzinfo=req.tz),
-                'yesterday': format_date(today - timedelta(days=1),
-                                         tzinfo=req.tz),
+                'today': user_time(req, format_date, today),
+                'yesterday': user_time(req, format_date, yesterday),
                 'precisedate': precisedate, 'precision': precision,
                 'events': [], 'filters': [],
-                'abbreviated_messages': self.abbreviated_messages}
+                'abbreviated_messages': self.abbreviated_messages,
+                'lastvisit': lastvisit}
 
         available_filters = []
         for event_provider in self.event_providers:
@@ -151,15 +164,16 @@ class TimelineModule(Component):
 
         # save the results of submitting the timeline form to the session
         if 'update' in req.args:
-            for filter in available_filters:
-                key = 'timeline.filter.%s' % filter[0]
-                if filter[0] in req.args:
+            for filter_ in available_filters:
+                key = 'timeline.filter.%s' % filter_[0]
+                if filter_[0] in req.args:
                     req.session[key] = '1'
                 elif key in req.session:
                     del req.session[key]
 
         stop = fromdate
-        start = stop - timedelta(days=daysback + 1)
+        start = to_datetime(stop.replace(tzinfo=None) -
+                            timedelta(days=daysback + 1), req.tz)
 
         # create author include and exclude sets
         include = set()
@@ -170,7 +184,7 @@ class TimelineModule(Component):
                 exclude.add(name)
             else:
                 include.add(name)
-        
+
         # gather all events for the given period of time
         events = []
         for provider in self.event_providers:
@@ -178,11 +192,11 @@ class TimelineModule(Component):
                 for event in provider.get_timeline_events(req, start, stop,
                                                           filters) or []:
                     # Check for 0.10 events
-                    author = (event[len(event) < 6 and 2 or 4] or '').lower()
+                    author = (event[2 if len(event) < 6 else 4] or '').lower()
                     if (not include or author in include) \
-                       and not author in exclude:
+                            and author not in exclude:
                         events.append(self._event_data(provider, event))
-            except Exception, e: # cope with a failure of that provider
+            except Exception, e:  # cope with a failure of that provider
                 self._provider_failure(e, req, provider, filters,
                                        [f[0] for f in available_filters])
 
@@ -192,25 +206,24 @@ class TimelineModule(Component):
             events = events[:maxrows]
 
         data['events'] = events
-        
 
         if format == 'rss':
-            # Get the email addresses of all known users
-            email_map = {}
-            if Chrome(self.env).show_email_addresses:
-                for username, name, email in self.env.get_known_users():
-                    if email:
-                        email_map[username] = email
-            data['email_map'] = email_map
-            rss_context = Context.from_request(req, absurls=True)
+            data['email_map'] = Chrome(self.env).get_email_map()
+            rss_context = web_context(req, absurls=True)
             rss_context.set_hints(wiki_flavor='html', shorten_lines=False)
             data['context'] = rss_context
             return 'timeline.rss', data, 'application/rss+xml'
         else:
-            req.session['timeline.daysback'] = daysback
-            req.session['timeline.authors'] = authors
-            html_context = Context.from_request(req)
-            html_context.set_hints(wiki_flavor='oneliner', 
+            req.session.set('timeline.daysback', daysback,
+                            self.default_daysback)
+            req.session.set('timeline.authors', authors, '')
+            # store lastvisit
+            if events and not revisit:
+                lastviewed = to_utimestamp(events[0]['date'])
+                req.session['timeline.lastvisit'] = max(lastvisit, lastviewed)
+                req.session['timeline.nextlastvisit'] = lastvisit
+            html_context = web_context(req)
+            html_context.set_hints(wiki_flavor='oneliner',
                                    shorten_lines=self.abbreviated_messages)
             data['context'] = html_context
 
@@ -218,29 +231,34 @@ class TimelineModule(Component):
         rss_href = req.href.timeline([(f, 'on') for f in filters],
                                      daysback=90, max=50, authors=authors,
                                      format='rss')
-        add_link(req, 'alternate', rss_href, _('RSS Feed'),
+        add_link(req, 'alternate', auth_link(req, rss_href), _('RSS Feed'),
                  'application/rss+xml', 'rss')
+        Chrome(self.env).add_jquery_ui(req)
 
         for filter_ in available_filters:
             data['filters'].append({'name': filter_[0], 'label': filter_[1],
                                     'enabled': filter_[0] in filters})
 
         # Navigation to the previous/next period of 'daysback' days
-        previous_start = format_date(fromdate - timedelta(days=daysback+1),
+        previous_start = fromdate.replace(tzinfo=None) - \
+                         timedelta(days=daysback + 1)
+        previous_start = format_date(to_datetime(previous_start, req.tz),
                                      format='%Y-%m-%d', tzinfo=req.tz)
         add_link(req, 'prev', req.href.timeline(from_=previous_start,
                                                 authors=authors,
                                                 daysback=daysback),
-                 _('Previous Period'))
+                 _("Previous Period"))
         if today - fromdate > timedelta(days=0):
-            next_start = format_date(fromdate + timedelta(days=daysback+1),
+            next_start = fromdate.replace(tzinfo=None) + \
+                         timedelta(days=daysback + 1)
+            next_start = format_date(to_datetime(next_start, req.tz),
                                      format='%Y-%m-%d', tzinfo=req.tz)
             add_link(req, 'next', req.href.timeline(from_=next_start,
                                                     authors=authors,
                                                     daysback=daysback),
-                     _('Next Period'))
-        prevnext_nav(req, _('Previous Period'), _('Next Period'))
-        
+                     _("Next Period"))
+        prevnext_nav(req, _("Previous Period"), _("Next Period"))
+
         return 'timeline.html', data, None
 
     # ITemplateProvider methods
@@ -255,13 +273,36 @@ class TimelineModule(Component):
 
     def pre_process_request(self, req, handler):
         return handler
-    
+
     def post_process_request(self, req, template, data, content_type):
         if data:
+            def pretty_dateinfo(date, format=None, dateonly=False):
+                absolute = user_time(req, format_datetime, date)
+                relative = pretty_timedelta(date)
+                if not format:
+                    format = req.session.get('dateinfo',
+                                 Chrome(self.env).default_dateinfo_format)
+                if format == 'absolute':
+                    if dateonly:
+                        label = absolute
+                    elif req.lc_time == 'iso8601':
+                        label = _("at %(iso8601)s", iso8601=absolute)
+                    else:
+                        label = _("on %(date)s at %(time)s",
+                                  date=user_time(req, format_date, date),
+                                  time=user_time(req, format_time, date))
+                    title = _("See timeline %(relativetime)s ago",
+                              relativetime=relative)
+                else:
+                    label = _("%(relativetime)s ago", relativetime=relative) \
+                            if not dateonly else relative
+                    title = _("See timeline at %(absolutetime)s",
+                              absolutetime=absolute)
+                return self.get_timeline_link(req, date, label,
+                                              precision='second', title=title)
             def dateinfo(date):
-                return self.get_timeline_link(req, date,
-                                              pretty_timedelta(date),
-                                              precision='second')
+                return pretty_dateinfo(date, format='relative', dateonly=True)
+            data['pretty_dateinfo'] = pretty_dateinfo
             data['dateinfo'] = dateinfo
         return template, data, content_type
 
@@ -272,8 +313,9 @@ class TimelineModule(Component):
 
     def get_link_resolvers(self):
         def link_resolver(formatter, ns, target, label):
+            path, query, fragment = split_url_into_path_query_fragment(target)
             precision = None
-            time = target.split("T", 1)
+            time = path.split("T", 1)
             if len(time) > 1:
                 time = time[1].split("Z")[0]
                 if len(time) >= 6:
@@ -283,25 +325,24 @@ class TimelineModule(Component):
                 elif len(time) >= 2:
                     precision = 'hours'
             try:
-                return self.get_timeline_link(formatter.req,
-                                              parse_date(target, utc),
-                                              label, precision)
+                dt = parse_date(path, utc, locale='iso8601', hint='iso8601')
+                return self.get_timeline_link(formatter.req, dt, label,
+                                              precision, query, fragment)
             except TracError, e:
-                return tag.a(label, title=to_unicode(e.message),
+                return tag.a(label, title=to_unicode(e),
                              class_='timeline missing')
-        yield ('timeline', link_resolver)
+        yield 'timeline', link_resolver
 
     # Public methods
 
-    def get_timeline_link(self, req, date, label=None, precision='hours'):
-        iso_date = display_date = format_datetime(date, 'iso8601', req.tz)
-        fmt = req.session.get('datefmt')
-        if fmt and fmt != 'iso8601':
-            display_date = format_datetime(date, fmt, req.tz)
+    def get_timeline_link(self, req, date, label=None, precision='hours',
+                          query=None, fragment=None, title=None):
+        iso_date = format_datetime(date, 'iso8601', req.tz)
+        href = req.href.timeline(from_=iso_date, precision=precision)
         return tag.a(label or iso_date, class_='timeline',
-                     title=_("%(date)s in Timeline", date=display_date),
-                     href=req.href.timeline(from_=iso_date,
-                                            precision=precision))
+                     title=title or _("See timeline at %(absolutetime)s",
+                                      absolutetime=iso_date),
+                     href=concat_path_query_fragment(href, query, fragment))
 
     # Internal methods
 
@@ -332,7 +373,7 @@ class TimelineModule(Component):
         At the same time, the message will contain a link to the timeline
         without the filters corresponding to the guilty event provider `ep`.
         """
-        self.log.error('Timeline event provider failed: %s', 
+        self.log.error("Timeline event provider failed: %s",
                        exception_to_unicode(exc, traceback=True))
 
         ep_kinds = dict((f[0], f[1])
@@ -341,9 +382,9 @@ class TimelineModule(Component):
         current_filters = set(current_filters)
         other_filters = set(current_filters) - ep_filters
         if not other_filters:
-            other_filters = set(all_filters) -  ep_filters
-        args = [(a, req.args.get(a)) for a in ('from', 'format', 'max',
-                                               'daysback')]
+            other_filters = set(all_filters) - ep_filters
+        args = [(a, req.args.get(a))
+                for a in ('from', 'format', 'max', 'daysback')]
         href = req.href.timeline(args + [(f, 'on') for f in other_filters])
         # TRANSLATOR: ...want to see the 'other kinds of events' from... (link)
         other_events = tag.a(_('other kinds of events'), href=href)
@@ -351,9 +392,9 @@ class TimelineModule(Component):
             tag.p(tag_("Event provider %(name)s failed for filters "
                        "%(kinds)s: ",
                        name=tag.tt(ep.__class__.__name__),
-                       kinds=', '.join('"%s"' % ep_kinds[f] for f in 
+                       kinds=', '.join('"%s"' % ep_kinds[f] for f in
                                        current_filters & ep_filters)),
-                  tag.b(exception_to_unicode(exc)), class_='message'),
+                  tag.strong(exception_to_unicode(exc)), class_='message'),
             tag.p(tag_("You may want to see the %(other_events)s from the "
                        "Timeline or notify your Trac administrator about the "
                        "error (detailed information was written to the log).",
