@@ -435,7 +435,7 @@ def encode_pair(name, value):
         s += struct.pack('!L', valueLength | 0x80000000L)
 
     return s + name + value
-    
+
 class Record(object):
     """
     A FastCGI Record.
@@ -450,6 +450,7 @@ class Record(object):
         self.paddingLength = 0
         self.contentData = ''
 
+    @staticmethod
     def _recvall(sock, length):
         """
         Attempts to receive length bytes from a socket, blocking if necessary.
@@ -473,7 +474,6 @@ class Record(object):
             recvLen += dataLen
             length -= dataLen
         return ''.join(dataList), recvLen
-    _recvall = staticmethod(_recvall)
 
     def read(self, sock):
         """Read and decode a Record from a socket."""
@@ -484,7 +484,7 @@ class Record(object):
 
         if length < FCGI_HEADER_LEN:
             raise EOFError
-        
+
         self.version, self.type, self.requestId, self.contentLength, \
                       self.paddingLength = struct.unpack(FCGI_Header, header)
 
@@ -492,7 +492,7 @@ class Record(object):
                              'contentLength = %d' %
                              (sock.fileno(), self.type, self.requestId,
                               self.contentLength))
-        
+
         if self.contentLength:
             try:
                 self.contentData, length = self._recvall(sock,
@@ -509,6 +509,7 @@ class Record(object):
             except:
                 raise EOFError
 
+    @staticmethod
     def _sendall(sock, data):
         """
         Writes data to a socket and does not return until all the data is sent.
@@ -525,7 +526,6 @@ class Record(object):
                     raise
             data = data[sent:]
             length -= sent
-    _sendall = staticmethod(_sendall)
 
     def write(self, sock):
         """Encode and write a Record to a socket."""
@@ -544,7 +544,7 @@ class Record(object):
             self._sendall(sock, self.contentData)
         if self.paddingLength:
             self._sendall(sock, '\x00'*self.paddingLength)
-            
+
 class Request(object):
     """
     Represents a single FastCGI request.
@@ -579,16 +579,16 @@ class Request(object):
         if __debug__: _debug(1, 'protocolStatus = %d, appStatus = %d' %
                              (protocolStatus, appStatus))
 
-        try: 
-            self._flush() 
-            self._end(appStatus, protocolStatus) 
-        except socket.error, e: 
-            if e[0] != errno.EPIPE: 
-                raise 
+        try:
+            self._flush()
+            self._end(appStatus, protocolStatus)
+        except socket.error, e:
+            if e[0] != errno.EPIPE:
+                raise
 
     def _end(self, appStatus=0L, protocolStatus=FCGI_REQUEST_COMPLETE):
         self._conn.end_request(self, appStatus, protocolStatus)
-        
+
     def _flush(self):
         self.stdout.close()
         self.stderr.close()
@@ -601,14 +601,14 @@ class CGIRequest(Request):
         self.role = FCGI_RESPONDER
         self.flags = 0
         self.aborted = False
-        
+
         self.server = server
         self.params = dict(os.environ)
         self.stdin = sys.stdin
         self.stdout = StdoutWrapper(sys.stdout) # Oh, the humanity!
         self.stderr = sys.stderr
         self.data = StringIO.StringIO()
-        
+
     def _end(self, appStatus=0L, protocolStatus=FCGI_REQUEST_COMPLETE):
         sys.exit(appStatus)
 
@@ -649,7 +649,7 @@ class Connection(object):
         except:
             pass
         self._sock.close()
-        
+
     def run(self):
         """Begin processing data from the socket."""
         self._keepGoing = True
@@ -810,7 +810,7 @@ class Connection(object):
         outrec.contentData = struct.pack(FCGI_UnknownTypeBody, inrec.type)
         outrec.contentLength = FCGI_UnknownTypeBody_LEN
         self.writeRecord(outrec)
-        
+
 class MultiplexedConnection(Connection):
     """
     A version of Connection capable of handling multiple requests
@@ -837,7 +837,7 @@ class MultiplexedConnection(Connection):
         self._lock.release()
 
         super(MultiplexedConnection, self)._cleanupSocket()
-        
+
     def writeRecord(self, rec):
         # Must use locking to prevent intermingling of Records from different
         # threads.
@@ -896,7 +896,7 @@ class MultiplexedConnection(Connection):
             super(MultiplexedConnection, self)._do_data(inrec)
         finally:
             self._lock.release()
-        
+
 class Server(object):
     """
     The FastCGI server.
@@ -1042,7 +1042,7 @@ class Server(object):
     def _restoreSignalHandlers(self):
         for signum,handler in self._oldSIGs:
             signal.signal(signum, handler)
-        
+
     def _hupHandler(self, signum, frame):
         self._hupReceived = True
         self._keepGoing = False
@@ -1150,7 +1150,7 @@ class WSGIServer(Server):
 
         Set multithreaded to False if your application is not MT-safe.
         """
-        if kw.has_key('handler'):
+        if 'handler' in kw:
             del kw['handler'] # Doesn't make sense to let this through
         super(WSGIServer, self).__init__(**kw)
 
@@ -1278,9 +1278,9 @@ class WSGIServer(Server):
 
     def _sanitizeEnv(self, environ):
         """Ensure certain values are present, if required by WSGI."""
-        if not environ.has_key('SCRIPT_NAME'):
+        if 'SCRIPT_NAME' not in environ:
             environ['SCRIPT_NAME'] = ''
-        if not environ.has_key('PATH_INFO'):
+        if 'PATH_INFO' not in environ:
             environ['PATH_INFO'] = ''
 
         # If any of these are missing, it probably signifies a broken
@@ -1289,12 +1289,12 @@ class WSGIServer(Server):
                              ('SERVER_NAME', 'localhost'),
                              ('SERVER_PORT', '80'),
                              ('SERVER_PROTOCOL', 'HTTP/1.0')]:
-            if not environ.has_key(name):
+            if name not in environ:
                 environ['wsgi.errors'].write('%s: missing FastCGI param %s '
                                              'required by WSGI!\n' %
                                              (self.__class__.__name__, name))
                 environ[name] = default
-            
+
 if __name__ == '__main__':
     def test_app(environ, start_response):
         """Probably not the most efficient example."""
