@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # Copyright (C) 2005-2009 Edgewall Software
-# Copyright (C) 2005-2006 Christian Boos <cboos@neuf.fr>
+# Copyright (C) 2005-2006 Christian Boos <cboos@edgewall.org>
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -12,15 +12,16 @@
 # individuals. For the exact contribution history, see the revision
 # history and logs, available at http://trac.edgewall.org/log/.
 #
-# Author: Christian Boos <cboos@neuf.fr>
+# Author: Christian Boos <cboos@edgewall.org>
 
 import re
 
 from genshi.builder import tag
 
 from trac.cache import cached
+from trac.config import ConfigSection
 from trac.core import *
-from trac.util.translation import _
+from trac.util.translation import _, N_
 from trac.wiki.api import IWikiChangeListener, IWikiMacroProvider
 from trac.wiki.parser import WikiParser
 from trac.wiki.formatter import split_url_into_path_query_fragment
@@ -31,6 +32,21 @@ class InterWikiMap(Component):
 
     implements(IWikiChangeListener, IWikiMacroProvider)
 
+    interwiki_section = ConfigSection('interwiki',
+        """Every option in the `[interwiki]` section defines one InterWiki
+        prefix. The option name defines the prefix. The option value defines
+        the URL, optionally followed by a description separated from the URL
+        by whitespace. Parametric URLs are supported as well.
+
+        '''Example:'''
+        {{{
+        [interwiki]
+        MeatBall = http://www.usemod.com/cgi-bin/mb.pl?
+        PEP = http://www.python.org/peps/pep-$1.html Python Enhancement Proposal $1
+        tsvn = tsvn: Interact with TortoiseSvn
+        }}}
+        """)
+
     _page_name = 'InterMapTxt'
     _interwiki_re = re.compile(r"(%s)[ \t]+([^ \t]+)(?:[ \t]+#(.*))?" %
                                WikiParser.LINK_SCHEME, re.UNICODE)
@@ -52,7 +68,7 @@ class InterWikiMap(Component):
         """Replace "$1" by the first args, "$2" by the second, etc."""
         def setarg(match):
             num = int(match.group()[1:])
-            return 0 < num <= len(args) and args[num-1] or ''
+            return args[num - 1] if 0 < num <= len(args) else ''
         return re.sub(InterWikiMap._argspec_re, setarg, txt)
 
     def _expand_or_append(self, txt, args):
@@ -60,16 +76,16 @@ class InterWikiMap(Component):
         if not args:
             return txt
         expanded = self._expand(txt, args)
-        return expanded == txt and txt + args[0] or expanded
+        return txt + args[0] if expanded == txt else expanded
 
     def url(self, ns, target):
         """Return `(url, title)` for the given InterWiki `ns`.
-        
+
         Expand the colon-separated `target` arguments.
         """
         ns, url, title = self[ns]
-        maxargnum = max([0]+[int(a[1:]) for a in
-                             re.findall(InterWikiMap._argspec_re, url)])
+        maxargnum = max([0] + [int(a[1:]) for a in
+                               re.findall(InterWikiMap._argspec_re, url)])
         target, query, fragment = split_url_into_path_query_fragment(target)
         if maxargnum > 0:
             args = target.split(':', (maxargnum - 1))
@@ -78,7 +94,7 @@ class InterWikiMap(Component):
         url = self._expand_or_append(url, args)
         ntarget, nquery, nfragment = split_url_into_path_query_fragment(url)
         if query and nquery:
-            nquery = '%s&%s' % (nquery, query[1:]) 
+            nquery = '%s&%s' % (nquery, query[1:])
         else:
             nquery = nquery or query
         nfragment = fragment or nfragment # user provided takes precedence
@@ -95,7 +111,7 @@ class InterWikiMap(Component):
         if page.name == InterWikiMap._page_name:
             del self.interwiki_map
 
-    def wiki_page_changed(self, page, version, t, comment, author, ipnr):
+    def wiki_page_changed(self, page, version, t, comment, author):
         if page.name == InterWikiMap._page_name:
             del self.interwiki_map
 
@@ -109,7 +125,7 @@ class InterWikiMap(Component):
 
     @cached
     def interwiki_map(self, db):
-        """Map from upper-cased namespaces to (namespace, prefix, title) 
+        """Map from upper-cased namespaces to (namespace, prefix, title)
         values.
         """
         from trac.wiki.model import WikiPage
@@ -125,10 +141,16 @@ class InterWikiMap(Component):
                     if m:
                         prefix, url, title = m.groups()
                         url = url.strip()
-                        title = title and title.strip() or prefix
+                        title = title.strip() if title else prefix
                         map[prefix.upper()] = (prefix, url, title)
             elif line.startswith('----'):
                 in_map = True
+        for prefix, value in self.interwiki_section.options():
+            value = value.split(None, 1)
+            if value:
+                url = value[0].strip()
+                title = value[1].strip() if len(value) > 1 else prefix
+                map[prefix.upper()] = (prefix, url, title)
         return map
 
     # IWikiMacroProvider methods
@@ -136,8 +158,10 @@ class InterWikiMap(Component):
     def get_macros(self):
         yield 'InterWiki'
 
-    def get_macro_description(self, name): 
-        return "Provide a description list for the known InterWiki prefixes."
+    def get_macro_description(self, name):
+        return 'messages', \
+               N_("Provide a description list for the known InterWiki "
+                  "prefixes.")
 
     def expand_macro(self, formatter, name, content):
         interwikis = []
@@ -146,10 +170,10 @@ class InterWikiMap(Component):
             interwikis.append({
                 'prefix': prefix, 'url': url, 'title': title,
                 'rc_url': self._expand_or_append(url, ['RecentChanges']),
-                'description': title == prefix and url or title})
+                'description': url if title == prefix else title})
 
-        return tag.table(tag.tr(tag.th(tag.em("Prefix")),
-                                tag.th(tag.em("Site"))),
+        return tag.table(tag.tr(tag.th(tag.em(_("Prefix"))),
+                                tag.th(tag.em(_("Site")))),
                          [tag.tr(tag.td(tag.a(w['prefix'], href=w['rc_url'])),
                                  tag.td(tag.a(w['description'],
                                               href=w['url'])))
