@@ -14,7 +14,7 @@
 # This plugin was based on the contrib/trac-post-commit-hook script, which
 # had the following copyright notice:
 # ----------------------------------------------------------------------------
-# Copyright (c) 2004 Stephen Hansen 
+# Copyright (c) 2004 Stephen Hansen
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to
@@ -24,7 +24,7 @@
 # furnished to do so, subject to the following conditions:
 #
 #   The above copyright notice and this permission notice shall be included in
-#   all copies or substantial portions of the Software. 
+#   all copies or substantial portions of the Software.
 #
 # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
@@ -35,6 +35,8 @@
 # IN THE SOFTWARE.
 # ----------------------------------------------------------------------------
 
+from __future__ import with_statement
+
 from datetime import datetime
 import re
 
@@ -46,9 +48,9 @@ from trac.perm import PermissionCache
 from trac.resource import Resource
 from trac.ticket import Ticket
 from trac.ticket.notification import TicketNotifyEmail
-from trac.util.compat import any
 from trac.util.datefmt import utc
 from trac.util.text import exception_to_unicode
+from trac.util.translation import _, cleandoc_
 from trac.versioncontrol import IRepositoryChangeListener, RepositoryManager
 from trac.versioncontrol.web_ui.changeset import ChangesetModule
 from trac.wiki.formatter import format_to_html
@@ -57,97 +59,97 @@ from trac.wiki.macros import WikiMacroBase
 
 class CommitTicketUpdater(Component):
     """Update tickets based on commit messages.
-    
+
     This component hooks into changeset notifications and searches commit
     messages for text in the form of:
     {{{
     command #1
     command #1, #2
-    command #1 & #2 
+    command #1 & #2
     command #1 and #2
     }}}
-    
+
     Instead of the short-hand syntax "#1", "ticket:1" can be used as well,
     e.g.:
     {{{
     command ticket:1
     command ticket:1, ticket:2
-    command ticket:1 & ticket:2 
+    command ticket:1 & ticket:2
     command ticket:1 and ticket:2
     }}}
-    
+
     In addition, the ':' character can be omitted and issue or bug can be used
     instead of ticket.
-    
+
     You can have more than one command in a message. The following commands
     are supported. There is more than one spelling for each command, to make
     this as user-friendly as possible.
-    
+
       close, closed, closes, fix, fixed, fixes::
         The specified tickets are closed, and the commit message is added to
         them as a comment.
-    
+
       references, refs, addresses, re, see::
         The specified tickets are left in their current status, and the commit
-        message is added to them as a comment. 
-    
+        message is added to them as a comment.
+
     A fairly complicated example of what you can do is with a commit message
     of:
-    
+
         Changed blah and foo to do this or that. Fixes #10 and #12,
         and refs #12.
-    
+
     This will close #10 and #12, and add a note to #12.
     """
-    
+
     implements(IRepositoryChangeListener)
-    
+
     envelope = Option('ticket', 'commit_ticket_update_envelope', '',
         """Require commands to be enclosed in an envelope.
-        
+
         Must be empty or contain two characters. For example, if set to "[]",
         then commands must be in the form of [closes #4].""")
-    
+
     commands_close = Option('ticket', 'commit_ticket_update_commands.close',
         'close closed closes fix fixed fixes',
         """Commands that close tickets, as a space-separated list.""")
-    
+
     commands_refs = Option('ticket', 'commit_ticket_update_commands.refs',
         'addresses re references refs see',
         """Commands that add a reference, as a space-separated list.
-        
+
         If set to the special value <ALL>, all tickets referenced by the
         message will get a reference to the changeset.""")
-    
+
     check_perms = BoolOption('ticket', 'commit_ticket_update_check_perms',
         'true',
         """Check that the committer has permission to perform the requested
         operations on the referenced tickets.
-        
+
         This requires that the user names be the same for Trac and repository
         operations.""")
 
     notify = BoolOption('ticket', 'commit_ticket_update_notify', 'true',
         """Send ticket change notification when updating a ticket.""")
-    
+
     ticket_prefix = '(?:#|(?:ticket|issue|bug)[: ]?)'
     ticket_reference = ticket_prefix + '[0-9]+'
     ticket_command = (r'(?P<action>[A-Za-z]*)\s*.?\s*'
                       r'(?P<ticket>%s(?:(?:[, &]*|[ ]?and[ ]?)%s)*)' %
                       (ticket_reference, ticket_reference))
-    
+
     @property
     def command_re(self):
         (begin, end) = (re.escape(self.envelope[0:1]),
                         re.escape(self.envelope[1:2]))
         return re.compile(begin + self.ticket_command + end)
-    
+
     ticket_re = re.compile(ticket_prefix + '([0-9]+)')
-    
+
     _last_cset_id = None
-    
+
     # IRepositoryChangeListener methods
-    
+
     def changeset_added(self, repos, changeset):
         if self._is_duplicate(changeset):
             return
@@ -155,7 +157,7 @@ class CommitTicketUpdater(Component):
         comment = self.make_ticket_comment(repos, changeset)
         self._update_tickets(tickets, changeset, comment,
                              datetime.now(utc))
-    
+
     def changeset_modified(self, repos, changeset, old_changeset):
         if self._is_duplicate(changeset):
             return
@@ -168,7 +170,7 @@ class CommitTicketUpdater(Component):
         comment = self.make_ticket_comment(repos, changeset)
         self._update_tickets(tickets, changeset, comment,
                              datetime.now(utc))
-    
+
     def _is_duplicate(self, changeset):
         # Avoid duplicate changes with multiple scoped repositories
         cset_id = (changeset.rev, changeset.message, changeset.author,
@@ -177,13 +179,14 @@ class CommitTicketUpdater(Component):
             self._last_cset_id = cset_id
             return False
         return True
-        
+
     def _parse_message(self, message):
         """Parse the commit message and return the ticket references."""
-        cmd_groups = self.command_re.findall(message)
+        cmd_groups = self.command_re.finditer(message)
         functions = self._get_functions()
         tickets = {}
-        for cmd, tkts in cmd_groups:
+        for m in cmd_groups:
+            cmd, tkts = m.group('action', 'ticket')
             func = functions.get(cmd.lower())
             if not func and self.commands_refs.strip() == '<ALL>':
                 func = self.cmd_refs
@@ -191,49 +194,56 @@ class CommitTicketUpdater(Component):
                 for tkt_id in self.ticket_re.findall(tkts):
                     tickets.setdefault(int(tkt_id), []).append(func)
         return tickets
-    
+
     def make_ticket_comment(self, repos, changeset):
         """Create the ticket comment from the changeset data."""
-        revstring = str(changeset.rev)
+        rev = changeset.rev
+        revstring = str(rev)
+        drev = str(repos.display_rev(rev))
         if repos.reponame:
             revstring += '/' + repos.reponame
+            drev += '/' + repos.reponame
         return """\
-In [changeset:%s]:
+In [changeset:"%s" %s]:
 {{{
 #!CommitTicketReference repository="%s" revision="%s"
 %s
-}}}""" % (revstring, repos.reponame, changeset.rev, changeset.message.strip())
-        
+}}}""" % (revstring, drev, repos.reponame, rev, changeset.message.strip())
+
     def _update_tickets(self, tickets, changeset, comment, date):
         """Update the tickets with the given comment."""
-        perm = PermissionCache(self.env, changeset.author)
+        authname = self._authname(changeset)
+        perm = PermissionCache(self.env, authname)
         for tkt_id, cmds in tickets.iteritems():
             try:
                 self.log.debug("Updating ticket #%d", tkt_id)
-                ticket = [None]
-                @self.env.with_transaction()
-                def do_update(db):
-                    ticket[0] = Ticket(self.env, tkt_id, db)
+                save = False
+                with self.env.db_transaction:
+                    ticket = Ticket(self.env, tkt_id)
+                    ticket_perm = perm(ticket.resource)
                     for cmd in cmds:
-                        cmd(ticket[0], changeset, perm(ticket[0].resource))
-                    ticket[0].save_changes(changeset.author, comment, date, db)
-                self._notify(ticket[0], date)
+                        if cmd(ticket, changeset, ticket_perm) is not False:
+                            save = True
+                    if save:
+                        ticket.save_changes(authname, comment, date)
+                if save:
+                    self._notify(ticket, date)
             except Exception, e:
                 self.log.error("Unexpected error while processing ticket "
                                "#%s: %s", tkt_id, exception_to_unicode(e))
-    
+
     def _notify(self, ticket, date):
         """Send a ticket update notification."""
         if not self.notify:
             return
+        tn = TicketNotifyEmail(self.env)
         try:
-            tn = TicketNotifyEmail(self.env)
             tn.notify(ticket, newticket=False, modtime=date)
         except Exception, e:
             self.log.error("Failure sending notification on change to "
                            "ticket #%s: %s", ticket.id,
                            exception_to_unicode(e))
-    
+
     def _get_functions(self):
         """Create a mapping from commands to command functions."""
         functions = {}
@@ -244,21 +254,38 @@ In [changeset:%s]:
             for cmd in getattr(self, 'commands_' + each[4:], '').split():
                 functions[cmd] = func
         return functions
-    
+
+    def _authname(self, changeset):
+        return changeset.author.lower() \
+               if self.env.config.getbool('trac', 'ignore_auth_case') \
+               else changeset.author
+
+    # Command-specific behavior
+    # The ticket isn't updated if all extracted commands return False.
+
     def cmd_close(self, ticket, changeset, perm):
-        if not self.check_perms or 'TICKET_MODIFY' in perm:
-            ticket['status'] = 'closed'
-            ticket['resolution'] = 'fixed'
-            if not ticket['owner']:
-                ticket['owner'] = changeset.author
+        authname = self._authname(changeset)
+        if self.check_perms and not 'TICKET_MODIFY' in perm:
+            self.log.info("%s doesn't have TICKET_MODIFY permission for #%d",
+                          authname, ticket.id)
+            return False
+        ticket['status'] = 'closed'
+        ticket['resolution'] = 'fixed'
+        if not ticket['owner']:
+            ticket['owner'] = authname
 
     def cmd_refs(self, ticket, changeset, perm):
-        pass
+        if self.check_perms and not 'TICKET_APPEND' in perm:
+            self.log.info("%s doesn't have TICKET_APPEND permission for #%d",
+                          self._authname(changeset), ticket.id)
+            return False
 
 
 class CommitTicketReferenceMacro(WikiMacroBase):
+    _domain = 'messages'
+    _description = cleandoc_(
     """Insert a changeset message into the output.
-    
+
     This macro must be called using wiki processor syntax as follows:
     {{{
     {{{
@@ -268,8 +295,8 @@ class CommitTicketReferenceMacro(WikiMacroBase):
     where the arguments are the following:
      - `repository`: the repository containing the changeset
      - `revision`: the revision of the desired changeset
-    """
-    
+    """)
+
     def expand_macro(self, formatter, name, content, args={}):
         reponame = args.get('repository') or ''
         rev = args.get('revision')
@@ -286,11 +313,11 @@ class CommitTicketReferenceMacro(WikiMacroBase):
             ticket_re = CommitTicketUpdater.ticket_re
             if not any(int(tkt_id) == int(formatter.context.resource.id)
                        for tkt_id in ticket_re.findall(message)):
-                return tag.p("(The changeset message doesn't reference this "
-                             "ticket)", class_='hint')
+                return tag.p(_("(The changeset message doesn't reference this "
+                               "ticket)"), class_='hint')
         if ChangesetModule(self.env).wiki_format_messages:
             return tag.div(format_to_html(self.env,
-                formatter.context('changeset', rev, parent=resource),
+                formatter.context.child('changeset', rev, parent=resource),
                 message, escape_newlines=True), class_='message')
         else:
             return tag.pre(message, class_='message')
